/* bfcm.c */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "bfcm.h"

static void check (const void * const ptr)
{
  if (NULL == ptr)
    {
      printf ("out of memory");
      exit (-1);
    }
}

/* retrun the BITS bits hash value for X */
static int hash_val (int x, const int bits)
{
  int res = 0;

  assert (bits>0);

  while (x > 0) {
    res ^= x;
    x = (unsigned long long)x >> bits;
  }
  return (res & ((1<<bits)-1));
}


static int rotate (int value, int total, int bits)
{
  int rtn;
  assert (total>0);
  assert (bits>=0);

  while (bits>=total)
    bits -= total;

  int mask = (1<<bits) - 1;
  rtn = (value << bits) & mask;
  rtn |= (((unsigned int) value) >> (total - bits));

  return rtn;
}


/* get the proper table size for NELEM elements */
static int get_table_size (int nelem)
{
   int tblg2;

   if (nelem >= 65)
      tblg2 = 6;
   else if (nelem >= 32)
      tblg2 = 5;
   else if (nelem >= 16)
      tblg2 = 4;
   else if (nelem >= 9)
      tblg2 = 3;
   else if (nelem >= 4)
      tblg2 = 2;
   else
      tblg2 = 1;

   return tblg2;
}




static void stream_finalize (bfcm *bfcm1)
{
   int int_bytes = ((bfcm1->l) / 32) + 1;

   int *rtn = realloc (bfcm1->stream, sizeof (int) * (int_bytes));

   if (!rtn) {
      fprintf (stderr, "Error in finalizing.\n");
      exit (-1);
   }

   bfcm1->stream = rtn;
   bfcm1->stream_length = int_bytes;
   bfcm1->r = int_bytes * 32;

   return;
}



/* use the context to predict left */
static int index_function_left (bfcm *bfcm1)
{
   int idx;
   int i;

   int *context = bfcm1->context;
   int delta = context[1] - context[0];

   idx = hash_val (delta, bfcm1->tbsizelg2);
   for (i = 1; i<bfcm1->order; i++) {
      int v;
      delta = context[i + 1] - context[i];
      v = hash_val (rotate (delta, bfcm1->tbsizelg2, i), bfcm1->tbsizelg2);
      idx ^= v;
   }

   return idx;
}


/* use the context to predict right */
static int index_function_right (bfcm *bfcm1)
{
   int idx;
   int i;

   int *context = bfcm1->context;
   int delta = context[0] - context[1];

   idx = hash_val (delta, bfcm1->tbsizelg2);
   for (i = 1; i<bfcm1->order; i++) {
      int v;
      delta = context[i] - context[i + 1];
      v = hash_val (rotate (delta, bfcm1->tbsizelg2, i), bfcm1->tbsizelg2);
      idx ^= v;
   }

   return idx;
}


static int context_update_forward (bfcm *bfcm1, const int value)
{
   int i;
   int *context = bfcm1->context;
   char order = bfcm1->order;
   int t;

   t = context[0];
   for (i = 1; i < order + 1; i++) {
      context[i - 1] = context[i];
   }
   context[order] = value;

   return t;
}

static int context_update_backward (bfcm *bfcm1, const int value)
{
   int i;
   int *context = bfcm1->context;
   char order = bfcm1->order;
   int t;

   t = context[order];
   for (i = order; i > 0; i--) {
      context[i] = context[i - 1];
   }
   context[0] = value;

   return t;
}


/* fetch BITS bits starting at OFFSET (bit) from the data stream DATA
 * store the value into VALUE as the result */
static int fetch_bits (int *data, int offset, int bits)
{
  int value = 0;
  int offset_byte = offset / 32;
  int r = offset - (offset_byte * 32);
  unsigned int selected;
  assert (bits == 1 || bits == 32);

  switch (bits)
  {
  case 1:
    if (r == 0)
    {
      selected = data[offset_byte];
      value = selected >> 31;
    }
    else
    {
      selected = data[offset_byte];
      value = (selected >> (31 - r)) & 0x1;
    }
    break;

  case 32:
      if (r == 0)
      {
         selected = data[offset_byte];
         value = selected;
      }
      else
      {
         selected = data[offset_byte];
         value = selected << r;

         selected = data[offset_byte + 1];
         value |= selected >> (32 - r);
         
      }
   }

  return value;
}


/* write value VALUE starting at OFFSET (bit) of the data stream DATA */
static void commit_bits (int *data, int value, int offset, int bits)
{
   assert (offset >= 0);
   int offset_byte = offset / 32;
   int r = offset - (offset_byte * 32);

   assert (bits == 1 || bits == 32);

   switch (bits)
   {
   case 1:
      assert (value == (value & 0x1));
      if (r == 0)
      {
         data[offset_byte] &= ~(1 << 31);
         data[offset_byte] |= value << 31;
      }
      else
      {
         data[offset_byte] &= ~(1 << (31 - r));
         data[offset_byte] |= value << (31 - r);
      }
      break;

   case 32:
      if (r == 0)
      {
         data[offset_byte] = value;
      }
      else
      {
         unsigned int value1 = ((unsigned int) value) >> r;
         data[offset_byte] &= ~(((1 << (32 - r)) - 1));
         data[offset_byte] |= value1;

         unsigned int value2 = (unsigned int) value << (32 - r);
         data[offset_byte + 1] &= (1 << (32 - r)) - 1;
         data[offset_byte + 1] |= value2;
      }
   }

   return;

}


/* extend the data stream by BYTES bytes */
static void extend_stream (bfcm *bfcm1, int bytes)
{
   if (bytes != 0)
   {
      int *rtn = realloc (bfcm1->stream, sizeof (int) * (bfcm1->stream_length + bytes));

      if (!rtn) {
         fprintf (stderr, "Error in finalizing.\n");
         exit (-1);
      }

      bfcm1->stream = rtn;
      bfcm1->stream_length = bfcm1->stream_length + bytes;
   }

   bfcm1->r = bfcm1->stream_length * 32;

   return;

}

/* use the context to predict the left value
 * used in forward_compress() and backward_uncompress() */
static int predict_left (bfcm *bfcm1, int predict)
{
  return bfcm1->context[0] - predict;
}


/* use the context to predict the left value
 * used in forward_uncompress() and backward_compress() */
static int predict_right (bfcm *bfcm1, int predict)
{
   return bfcm1->context[bfcm1->order] - predict;
}


/* Basic Operations
 * As described in TACO 05 */
static void forward_compress (bfcm *bfcm1, int value)
{
   int idx = index_function_left (bfcm1);
   assert (idx>=0 && idx<(bfcm1->table_size));

   int pred = predict_left (bfcm1, bfcm1->frtb[idx]);

   if (pred == value) {
      // hit1
      bfcm1->l += 1;
      commit_bits (bfcm1->stream, 1, bfcm1->l, 1);
   }
   else {
      // miss1
      bfcm1->l += 1;
      commit_bits (bfcm1->stream, bfcm1->frtb[idx], bfcm1->l, 32);

      bfcm1->l += 32;
      commit_bits (bfcm1->stream, 0, bfcm1->l, 1);

      bfcm1->frtb[idx] = predict_left (bfcm1, value);
   }


}

/* Basic Operations */
static void backward_compress (bfcm *bfcm1, int value)
{
   int idx = index_function_right (bfcm1);
   assert (idx>=0 && idx<(bfcm1->table_size));

   int pred = predict_right (bfcm1, bfcm1->bltb[idx]);

   if (pred == value) {
      // hit
      bfcm1->r -= 1;
      commit_bits (bfcm1->stream, 1, bfcm1->r, 1);
   }
   else {
      // miss
      bfcm1->r -= 32;
      commit_bits (bfcm1->stream, bfcm1->bltb[idx], bfcm1->r, 32);

      bfcm1->r -= 1;
      commit_bits (bfcm1->stream, 0, bfcm1->r, 1);

      bfcm1->bltb[idx] = predict_right (bfcm1, value);
   }

}


/* Basic Operations */
static int backward_uncompress (bfcm *bfcm1)
{
   char b;
   int v;
   int *stream = bfcm1->stream;

   b = fetch_bits (bfcm1->stream, bfcm1->l, 1);
   bfcm1->l -= 1;

   int index = index_function_left (bfcm1);
   assert (index>=0 && index<(bfcm1->table_size));

   int delta = bfcm1->frtb[index];
   int pred = predict_left (bfcm1, delta);

   if (b == 1)
   {
      v = pred;
   }
   else
   {
      int v1;
      v1 = fetch_bits (bfcm1->stream, bfcm1->l - 31, 32);
      bfcm1->l -= 32;
      bfcm1->frtb[index] = v1;
      v = pred;
   }

   return v;
}

/* Basic Operations */
static int forward_uncompress (bfcm *bfcm1)
{
   char b;
   int v;
   int *stream = bfcm1->stream;

   b = fetch_bits (bfcm1->stream, bfcm1->r, 1);
   bfcm1->r += 1;
   int index = index_function_right (bfcm1);
   assert (index>=0 && index<(bfcm1->table_size));

   int delta = bfcm1->bltb[index];
   int pred = predict_right (bfcm1, delta);

   if (b == 1)
   {
      v = pred;
   }
   else
   {
      int v1 = fetch_bits (bfcm1->stream, bfcm1->r, 32);
      bfcm1->r += 32;
      if (bfcm1->r >= bfcm1->stream_length * 32)
         fprintf (stderr, "r=%d\tlength*32=%d\n", bfcm1->r, bfcm1->stream_length);
      assert (bfcm1->r < bfcm1->stream_length * 32);

      bfcm1->bltb[index] = v1;
      v = pred;
   }

   return v;
}

/* This algorithm first forward compresses all the values into the bit stream.
 * However, Tmp is not yet bidirectional traversable. A backward traversal
 * will be performed later to determine the amount of additional space that
 * needs to be allocated. */
static void building_forward (bfcm *bfcm1)
{
   int *datain = bfcm1->datain;
   int nelem = bfcm1->nelem;
   int i;

   // preload the first (bfcm1->order) elements into the context
   for (i = 0; i < bfcm1->order + 1; i++)
   {
      bfcm1->context[i] = bfcm1->datain[i];
   }

   // forward compress pass
   for (i = 0; i < nelem; i ++)
   {
      int t;

      if (i + bfcm1->order + 1 >= nelem)
         t = context_update_forward (bfcm1, 0);
      else
         t = context_update_forward (bfcm1, datain[i + bfcm1->order + 1]);
      assert (t==datain[i]);

      int idx = index_function_left (bfcm1);
      assert (idx>=0 && idx<(bfcm1->table_size));

      int pred = predict_left (bfcm1, bfcm1->frtb[idx]);
      int value = datain[i];

      if (pred == value) {
         // hit

         if (bfcm1->l + 1 >= bfcm1->r)
         {
            // stream space is not enough, extend STREAM_GROWTH bytes
            extend_stream (bfcm1, STREAM_GROWTH);
         }

         commit_bits (bfcm1->stream, 1, bfcm1->l + 1, 1);
         bfcm1->l += 1;

         bfcm1->hit1 ++;
      }
      else {
         // miss
         if (bfcm1->l + 33 >= bfcm1->r)
         {
            // stream space is not enough, extend STREAM_GROWTH bytes
            extend_stream (bfcm1, STREAM_GROWTH);
         }

         commit_bits (bfcm1->stream, bfcm1->frtb[idx], bfcm1->l + 1, 32);
         commit_bits (bfcm1->stream, 0, bfcm1->l + 33, 1);

         bfcm1->l += 33;

         bfcm1->miss1 ++;

         bfcm1->frtb[idx] = predict_left (bfcm1, value);
      }
   }

   stream_finalize (bfcm1);

//   fprintf (stderr, "[building_forward] l=%d\tr=%d\n", bfcm1->l, bfcm1->r);

}


/* A backward traversal is performed to determine the amount of additional
 * space that needs to be allocated. */
static void building_backward (bfcm *bfcm1)
{
   int i;
   int j;
   int nelem = bfcm1 -> nelem;
   int *stream = bfcm1->stream;
   int *datain = bfcm1->datain;
   unsigned int extra = 0;
   unsigned int ll = bfcm1->l;

//   fprintf (stderr, "[building_backward] l=%d\tr=%d\n", bfcm1->l, bfcm1->r);

   for (i = nelem - 1; i >= 0; i--)
   {
      int v = backward_uncompress (bfcm1);
      int value = datain[i];

      assert (value==v);

      int t = context_update_backward (bfcm1, v);

      int idx = index_function_right (bfcm1);
      assert (idx>=0 && idx<(bfcm1->table_size));

      int pred = predict_right (bfcm1, bfcm1->bltb[idx]);

      if (pred == t) {
         // hit
         bfcm1->r -= 1;
         if (bfcm1->r < 0)
         {
            extra += 1;
         }
         bfcm1->hit2 ++;
      }
      else {
         // miss
         if (bfcm1->r - 33 <= bfcm1->l) {
            assert (bfcm1->l - (bfcm1->r - 33) + 1 > 0);

            extra += bfcm1->l - (bfcm1->r - 33) + 1;
            bfcm1->r = bfcm1->l + 1;
         }
         else {
            bfcm1->r -= 33;
         }
         bfcm1->miss2 ++;
         bfcm1->bltb[idx] = predict_right (bfcm1, t);
      }
   }

//   fprintf (stderr, "bfcm1->hit2=%d\tmiss2=%d\n", bfcm1->hit2, bfcm1->miss2);

   bfcm1->hit2 = 0;
   bfcm1->miss2 = 0;

   if (extra % 32 ==0)
      extra = extra / 32;
   else
      extra = extra / 32 + 1;

//   fprintf (stderr, "extra buffer=%d ints\n", extra);
   extend_stream (bfcm1, extra);

   memset (bfcm1->bltb, 0, sizeof(int) * bfcm1->table_size);
//   fprintf (stderr, "l=%d\tr=%d\n", bfcm1->l, bfcm1->r);


   // preload the first (bfcm1->order) elements into the context
   for (i = 0; i < bfcm1->order + 1; i++)
   {
      bfcm1->context[i] = 0;
   }

   // redo the backward compress to generate the final stream
   for (i = nelem - 1; i >= 0; i--)
   {
      int value = datain[i];
      int target = context_update_backward (bfcm1, value);

      int idx = index_function_right (bfcm1);
      assert (idx>=0 && idx<(bfcm1->table_size));

      int pred = predict_right (bfcm1, bfcm1->bltb[idx]);

      if (pred == target) {
         // hit
         bfcm1->r -= 1;
         assert (bfcm1->r >= 0);

         commit_bits (bfcm1->stream, 1, bfcm1->r, 1);
         int t = fetch_bits (bfcm1->stream, bfcm1->r, 1);
         assert (t==1);

         bfcm1->hit2 ++;
      }
      else {
         // miss
         bfcm1->r -= 32;
         assert (bfcm1->r >= 0);

         commit_bits (bfcm1->stream, bfcm1->bltb[idx], bfcm1->r, 32);
         int t = fetch_bits (bfcm1->stream, bfcm1->r, 32);
         assert (t==bfcm1->bltb[idx]);

         bfcm1->r -= 1;
         assert (bfcm1->r >= 0);

         commit_bits (bfcm1->stream, 0, bfcm1->r, 1);
         t = fetch_bits (bfcm1->stream, bfcm1->r, 1);
         assert (t==0);

         bfcm1->miss2 ++;
         bfcm1->bltb[idx] = predict_right (bfcm1, target);

      }
   }

   return;
}


static int step_forward (bfcm *bfcm1)
{
   int v = forward_uncompress (bfcm1);
   int t = context_update_forward (bfcm1, v);

   forward_compress (bfcm1, t);

   return t;
}


static int step_backward (bfcm *bfcm1)
{
   int v = backward_uncompress (bfcm1);
   int t = context_update_backward (bfcm1, v);

   backward_compress (bfcm1, t);

   return v;
}


static void validate_forward_traverse (bfcm *bfcm1)
{
   int i;
   int value;

   fprintf (stderr, "[validate_forward] l=%d\tr=%d... ", bfcm1->l, bfcm1->r);

   for (i = 0; i < bfcm1->nelem; i++)
   {
      value = step_forward (bfcm1);
      assert (value == bfcm1->datain[i]);
   }
   fprintf (stderr, "PASS\n");
}


static void validate_backward_traverse (bfcm *bfcm1)
{
   int i;
   int value;

   fprintf (stderr, "[validate_backward] l=%d\tr=%d... ", bfcm1->l, bfcm1->r);
   for (i = bfcm1->nelem - 1; i >= 0; i--)
   {
      value = step_backward (bfcm1);
      assert (value == bfcm1->datain[i]);
   }
   fprintf (stderr, "PASS\n");
}


/* validate the bfcm_get_element by accessing the value arbitrarily */
static void validate_get_element (bfcm *bfcm1)
{
   int i;
   int value;

   fprintf (stderr, "[validate_bfcm_get_element] ... ");
   for (i = 0; i < bfcm1->nelem / 2; i++) {
      bfcm_get_element (bfcm1, i, &value);
      assert (value == bfcm1->datain[i]);
   }

   for ( /* cont */; i >= bfcm1->nelem / 3; i--) {
      bfcm_get_element (bfcm1, i, &value);
      assert (value == bfcm1->datain[i]);
   }

   for ( /* cont */; i < bfcm1->nelem; i++) {
      bfcm_get_element (bfcm1, i, &value);
      assert (value == bfcm1->datain[i]);
   }

   fprintf (stderr, "PASS\n");
   return;
}

static void bfcm_allocate (bfcm *bfcm1)
{
   int *table;
   int *context;
   int table_size = bfcm1->table_size;

   // allocate space for frtb  
   table = (int *) malloc (sizeof(int) * table_size);
   check (table);
   memset (table, 0, sizeof(int) * table_size);
   bfcm1->frtb = table;

   // allocate space for bltb
   table = (int *) malloc (sizeof(int) * table_size);
   check (table);
   memset (table, 0, sizeof(int) * table_size);
   bfcm1->bltb = table;

   // allocate space for context
   context = (int *) malloc (sizeof (int) * (bfcm1->order + 1));
   check (context);
   memset (context, 0, sizeof (int) * (bfcm1->order + 1));
   bfcm1->context = context;

}

/* do the initialization, allocate the necessary space for stream and bfcm tables */
bfcm* bfcm_init (int *datain, int nelem, int order)
{
   int table_size;
   int tblg2;

   // we support only orders from 1 to 6
   if (order < 1 || order > 6)
      return NULL;

   // too few elements to compress
   if (nelem < order + 2)
      return NULL;

   bfcm* bfcm1 = (bfcm *) malloc (sizeof (bfcm));
   check (bfcm1);

   bfcm1->order = order;

   tblg2 = get_table_size (nelem);

   table_size = 1 << tblg2;
   bfcm1->table_size = table_size;
   bfcm1->tbsizelg2 = tblg2;

   bfcm1->datain = datain;
   bfcm1->nelem = nelem;

   bfcm1->hit1 = 0;
   bfcm1->miss1 = 0;

   bfcm1->hit2 = 0;
   bfcm1->miss2 = 0;

   // initial stream length, may grow dynamically
   bfcm1->stream = (int *) malloc (sizeof (int) * STREAM_GROWTH);
   check  (bfcm1->stream);
   bfcm1->stream_length = STREAM_GROWTH;

   bfcm_allocate (bfcm1);

   // current element index
   bfcm1->i = -1;

   // reset the stream boundary pointer
   bfcm1->l = -1;
   bfcm1->r = STREAM_GROWTH * 32;

   return bfcm1;
}


/* free the space allocated for bfcm stream and tables */
void bfcm_free (bfcm *bfcm1)
{
   assert (bfcm1);

   free (bfcm1->frtb);
   free (bfcm1->bltb);
   free (bfcm1->stream);
   free (bfcm1->context);

   check (bfcm1);
   free (bfcm1);

   return;
}


/* As described in TACO'05 pp.319
 * Preparing data stream for bidirectional traversal.
*/
int bfcm_compress (bfcm *bfcm1)
{
   // first pass, forward generate the fltb, allocate space for the data stream
   building_forward (bfcm1);
  // second pass, backward generate bltb and reserve enough space in the buffer
   building_backward (bfcm1);

   return 0;
}

/* validate the forward traverse, backward traverse and the get element methods */
int bfcm_validate (bfcm *bfcm1)
{
   fprintf (stderr, "\nbfcm_validate validating...\n");

   validate_forward_traverse (bfcm1);
   validate_backward_traverse (bfcm1);
   validate_get_element (bfcm1);

   // it will halt on the assert if any error occurs
   return 0;
}

/* return the total length for storing the bfcm compressed stream */
int bfcm_length (bfcm *bfcm1)
{
   return sizeof (int) * (bfcm1->stream_length + bfcm1->table_size * 2 + (bfcm1->order + 1)) + sizeof (bfcm);
}


/* write the whole bfcm structure along with its content to disk */
int bfcm_commit (bfcm *bfcm1, FILE *fout)
{
   if (fout == NULL)
   {
      fprintf (stderr, "bfcm_commit: file not open\n");
      return -1;
   }

   fwrite (bfcm1, sizeof (bfcm), 1, fout);

   fwrite (bfcm1->stream, sizeof (int), bfcm1->stream_length, fout);

   fwrite (bfcm1->frtb, sizeof (int), bfcm1->table_size, fout);
   fwrite (bfcm1->bltb, sizeof (int), bfcm1->table_size, fout);

   fwrite (bfcm1->context, sizeof (int), (bfcm1->order) + 1, fout);

   return 0;
}

bfcm *bfcm_loadin (FILE *fin)
{
   bfcm *bfcm1;
   if (fin == NULL)
   {
      fprintf (stderr, "bfcm_loadin: file not open\n");
      return NULL;
   }

   bfcm1 = (bfcm *) malloc (sizeof (bfcm));

   fread (bfcm1, sizeof (bfcm), 1, fin);

   bfcm_allocate (bfcm1);
   bfcm1->stream = (int *) malloc (sizeof (int) * bfcm1->stream_length);

   fread (bfcm1->stream, sizeof (int), bfcm1->stream_length, fin);

   fread (bfcm1->frtb, sizeof (int), bfcm1->table_size, fin);
   fread (bfcm1->bltb, sizeof (int), bfcm1->table_size, fin);

   fread (bfcm1->context, sizeof (int), (bfcm1->order) + 1, fin);

   return bfcm1;
}


/* get the element at INDEX, store the value in VALUE */
int bfcm_get_element (bfcm *bfcm1, int index, int *value)
{
   int ci;
   
   if (index < 0 || index >= bfcm1->nelem)
      return -1;

   ci = bfcm1->i;  // get current index, which is the last element in the left compressed stream

   if (ci == index) {
      // nothing to do, return current value
   }
   else if (ci < index) {
      int t1;
      while (ci < index) {
         t1 = step_forward (bfcm1);
         ci ++;
      }
      bfcm1->i = index;
      bfcm1->current = t1;
   }
   else {
      int t1;
      while (ci >= index) {
         t1 = step_backward (bfcm1);
         ci --;
      }
      t1 = step_forward (bfcm1);
      bfcm1->i = index;
      bfcm1->current = t1;
   }

   *value = bfcm1->current;

   return 0;   
}


